# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import re
from slugify import slugify


# Internal

def convert_table(table, prefix):
    """Convert high-level table name to database name.
    """
    return prefix + table


def restore_table(table, prefix):
    """Restore database table name to high-level name.
    """
    if table.startswith(prefix):
        return table.replace(prefix, '', 1)
    return None


def convert_schema(schema):
    """Convert JSONTableSchema schema to BigQuery schema.
    """

    # Mapping
    mapping = {
        'string': 'STRING',
        'integer': 'INTEGER',
        'number': 'FLOAT',
        'boolean': 'BOOLEAN',
        'datetime': 'TIMESTAMP',
    }

    # Schema
    fields = []
    for field in schema['fields']:
        try:
            ftype = mapping[field['type']]
        except KeyError:
            message = 'Type %s is not supported' % field['type']
            raise TypeError(message)
        mode = 'NULLABLE'
        if field.get('constraints', {}).get('required', False):
            mode = 'REQUIRED'
        resfield = {
            'name': convert_field_name(field['name']),
            'type': ftype,
            'mode': mode,
        }
        fields.append(resfield)
    schema = {'fields': fields}

    return schema


def restore_schema(schema):
    """Convert BigQuery schema to JSONTableSchema schema.
    """

    # Mapping
    mapping = {
        'STRING': 'string',
        'INTEGER': 'integer',
        'FLOAT': 'number',
        'BOOLEAN': 'boolean',
        'TIMESTAMP': 'datetime',
    }

    # Schema
    fields = []
    for field in schema['fields']:
        try:
            ftype = mapping[field['type']]
        except KeyError:
            message = 'Type %s is not supported' % field['type']
            raise TypeError(message)
        resfield = {
            'name': field['name'],
            'type': ftype,
        }
        if field.get('mode', 'NULLABLE') != 'NULLABLE':
            resfield['constraints'] = {'required': True}
        fields.append(resfield)
    schema = {'fields': fields}

    return schema


def convert_field_name(name):
    # Check https://cloud.google.com/bigquery/docs/reference/v2/tables for
    # reference
    MAX_LENGTH = 128
    VALID_NAME = '^[a-zA-Z_]\w{0,%d}$' % (MAX_LENGTH-1)

    if not re.match(VALID_NAME, name):
        name = slugify(name, separator='_')
        if not re.match('^[a-zA-Z_]', name):
            name = '_' + name

    return name[:MAX_LENGTH]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import six
import time
import jsontableschema
import unicodecsv as csv
from jsontableschema.model import SchemaModel
from apiclient.http import MediaIoBaseUpload

from . import mappers


# Module API

class Storage(object):
    """BigQuery Tabular Storage.

    Parameters
    ----------
    service: object
        Service object from API.
    project: str
        Project name.
    dataset: str
        Dataset name.
    prefix: str
        Prefix for all tables.

    """

    # Public

    def __init__(self, service, project, dataset, prefix=''):

        # Set attributes
        self.__service = service
        self.__project = project
        self.__dataset = dataset
        self.__prefix = prefix
        self.__tables = None

    def __repr__(self):

        # Template and format
        template = 'Storage <{service}/{project}-{dataset}>'
        text = template.format(
                service=self.__service,
                project=self.__project,
                dataset=self.__dataset)

        return text

    @property
    def tables(self):
        """Return list of storage's table names.
        """

        # No cached value
        if self.__tables is None:

            # Get response
            response = self.__service.tables().list(
                    projectId=self.__project,
                    datasetId=self.__dataset).execute()

            # Extract tables
            tables = []
            for table in response.get('tables', []):
                table = table['tableReference']['tableId']
                table = mappers.restore_table(table, self.__prefix)
                if table is not None:
                    tables.append(table)

            # Save to cache
            self.__tables = tables

        return self.__tables

    def check(self, table):
        """Return if table exists.
        """

        # Check existence
        existence = table in self.tables

        return existence

    def create(self, table, schema):
        """Create table by schema.

        Parameters
        ----------
        table: str/list
            Table name or list of table names.
        schema: dict/list
            JSONTableSchema schema or list of schemas.

        Raises
        ------
        RuntimeError
            If table already exists.

        """

        # Make lists
        tables = table
        if isinstance(table, six.string_types):
            tables = [table]
        schemas = schema
        if isinstance(schema, dict):
            schemas = [schema]

        # Iterate over tables/schemas
        for table, schema in zip(tables, schemas):

            # Check not existent
            if self.check(table):
                message = 'Table "%s" already exists' % table
                raise RuntimeError(message)

            # Validate schema
            jsontableschema.validate(schema)

            # Prepare job body
            schema = mappers.convert_schema(schema)
            table = mappers.convert_table(table, self.__prefix)
            body = {
                'tableReference': {
                    'projectId': self.__project,
                    'datasetId': self.__dataset,
                    'tableId': table,
                },
                'schema': schema,
            }

            # Make request
            self.__service.tables().insert(
                    projectId=self.__project,
                    datasetId=self.__dataset,
                    body=body).execute()

        # Remove tables cache
        self.__tables = None

    def delete(self, table):
        """Delete table.

        Parameters
        ----------
        table: str/list
            Table name or list of table names.

        Raises
        ------
        RuntimeError
            If table doesn't exist.

        """

        # Make lists
        tables = table
        if isinstance(table, six.string_types):
            tables = [table]

        # Iterater over tables
        for table in tables:

            # Check existent
            if not self.check(table):
                message = 'Table "%s" doesn\'t exist.' % self
                raise RuntimeError(message)

            # Make delete request
            table = mappers.convert_table(table, self.__prefix)
            self.__service.tables().delete(
                    projectId=self.__project,
                    datasetId=self.__dataset,
                    tableId=table).execute()

        # Remove tables cache
        self.__tables = None

    def describe(self, table):
        """Return table's JSONTableSchema schema.

        Parameters
        ----------
        table: str
            Table name.

        Returns
        -------
        dict
            JSONTableSchema schema.

        """

        # Get response
        table = mappers.convert_table(table, self.__prefix)
        response = self.__service.tables().get(
                projectId=self.__project,
                datasetId=self.__dataset,
                tableId=table).execute()

        # Get schema
        schema = response['schema']
        schema = mappers.restore_schema(schema)

        return schema

    def read(self, table):
        """Read data from table.

        Parameters
        ----------
        table: str
            Table name.

        Returns
        -------
        generator
            Data tuples generator.

        """

        # Get response
        schema = self.describe(table)
        model = SchemaModel(schema)
        table = mappers.convert_table(table, self.__prefix)
        response = self.__service.tabledata().list(
                projectId=self.__project,
                datasetId=self.__dataset,
                tableId=table).execute()

        # Yield data
        for row in response['rows']:
            row = tuple(field['v'] for field in row['f'])
            row = tuple(model.convert_row(*row, fail_fast=True))
            yield row

    def write(self, table, data):
        """Write data to table.

        Parameters
        ----------
        table: str
            Table name.
        data: list
            List of data tuples.

        """

        # Process data to byte stream csv
        schema = self.describe(table)
        model = SchemaModel(schema)
        bytes = io.BufferedRandom(io.BytesIO())
        writer = csv.writer(bytes, encoding='utf-8')
        for row in data:
            row = tuple(model.convert_row(*row, fail_fast=True))
            writer.writerow(row)
        bytes.seek(0)

        # Prepare job body
        table = mappers.convert_table(table, self.__prefix)
        body = {
            'configuration': {
                'load': {
                    'destinationTable': {
                        'projectId': self.__project,
                        'datasetId': self.__dataset,
                        'tableId': table
                    },
                    'sourceFormat': 'CSV',
                }
            }
        }

        # Prepare job media body
        mimetype = 'application/octet-stream'
        media_body = MediaIoBaseUpload(bytes, mimetype=mimetype)

        # Make request to Big Query
        response = self.__service.jobs().insert(
            projectId=self.__project,
            body=body,
            media_body=media_body).execute()
        self.__wait_response(response)

    # Private

    def __wait_response(self, response):

        # Get job instance
        job = self.__service.jobs().get(
            projectId=response['jobReference']['projectId'],
            jobId=response['jobReference']['jobId'])

        # Wait done
        while True:
            result = job.execute(num_retries=1)
            if result['status']['state'] == 'DONE':
                if result['status'].get('errors'):
                    errors = result['status']['errors']
                    message = '\n'.join(error['message'] for error in errors)
                    raise RuntimeError(message)
                break
            time.sleep(1)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import io
import six
import sys
import json
import runpy
import tempfile
import unittest

from examples import base
from examples import testing


class TestStorage(unittest.TestCase):

    # Tests

    def test_testing(self):

        # Run function
        tables, schema, data = base.run(
                testing.dataset, testing.prefix, testing.table,
                testing.schema, testing.data)

        # Assert values
        assert testing.table in tables
        assert schema == testing.schema
        assert data == testing.data



# -*- coding: utf-8 -*-
# pylama:skip=1
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from .storage import Storage



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
import unittest
from mock import MagicMock, patch, ANY
from importlib import import_module
module = import_module('jsontableschema_bigquery.storage')


class TestStorage(unittest.TestCase):

    # Helpers

    def setUp(self):

        # Mocks
        self.addCleanup(patch.stopall)
        self.service = MagicMock()
        self.project = 'project'
        self.dataset = 'dataset'
        self.prefix = 'prefix_'

        # Create storage
        self.storage = module.Storage(
                service=self.service,
                project=self.project,
                dataset=self.dataset,
                prefix=self.prefix)

    # Tests

    def test___repr__(self):

        # Assert values
        assert repr(self.storage)

    def test_tables(self):

        # Mocks
        self.service.tables.return_value.list.return_value.execute.return_value = {
            'tables': [
                {'tableReference': {'tableId': 'prefix_table1'}},
                {'tableReference': {'tableId': 'prefix_table2'}},
            ],
        }

        # Assert values
        assert self.storage.tables == ['table1', 'table2']

    def test_check(self):

        # Mocks
        self.service.tables.return_value.list.return_value.execute.return_value = {
            'tables': [
                {'tableReference': {'tableId': 'prefix_table1'}},
                {'tableReference': {'tableId': 'prefix_table2'}},
            ],
        }

        # Assert values
        assert self.storage.check('table1')
        assert self.storage.check('table2')
        assert not self.storage.check('table3')

    def test_create(self):
        pass

    def test_create_existent(self):

        # Mocks
        self.storage.check = MagicMock(return_value=True)

        # Assert raises
        with pytest.raises(RuntimeError):
            self.storage.create('table', 'schema')

    def test_delete(self):

        # Mocks
        self.storage.check = MagicMock(return_value=True)

        # Method call
        self.storage.delete('table')

        # Assert calls
        self.service.tables.return_value.delete.assert_called_with(
                projectId=self.project,
                datasetId=self.dataset,
                tableId='prefix_table')

    def test_delete_non_existent(self):

        # Assert exception
        with pytest.raises(RuntimeError):
           self.storage.delete('table')

    def test_describe(self):
        pass

    def test_read(self):
        pass

    def test_write(self):
        pass
